#include "main.hpp"
#include "Game.hpp"
#include "TextGame.hpp"
#include "TextGameReader.hpp"
#include "ManyRunsLogicInfo.hpp"

/**
 * To do:
 *  + The graph of turret upgrades
 *  + Better permutations?
 */

void stupidTest(){
    char field[MAX_FIELD_SIZE][MAX_FIELD_SIZE];
    
    strcpy( field[0] , "S###E" );
    strcpy( field[1] , "....." );
    
    /**
     * Field construction
     */
    Game *game = new Game( 2 , 5 , field );
    
        game->money = 100ULL;
    
    /**
     * Mob creation
     */
    Mob *spider = new Mob;
        
        spider->type = 0;
        spider->speed = 1;
        spider->health = 10;
    
    std::vector< RunningMob* > mobs;
    
    mobs.push_back( new RunningMob( spider ) );
    mobs.push_back( new RunningMob( spider ) );
    mobs.push_back( new RunningMob( spider ) );
    
    /**
     * Create the Turret
     */
    Turret *rocketLauncher = new Turret;
    
        rocketLauncher->rate = 5;
        rocketLauncher->range = 3;
        rocketLauncher->buy_price = 50;
        rocketLauncher->sell_price = 20;
        rocketLauncher->affect[0] = 1;
        rocketLauncher->name = "Rocket Launcher";
        rocketLauncher->type = TURRET_BATTLE;
    
    Turret *doubleDamage = new Turret;
    
        doubleDamage->range = 1;
        doubleDamage->buy_price = 50;
        doubleDamage->sell_price = 20;
        doubleDamage->affect[0] = 0.5;
        doubleDamage->name = "Double damage";
        doubleDamage->type = TURRET_MOD_DAMAGE;
        
    Turret *doubleRate = new Turret;
    
        doubleRate->range = 1;
        doubleRate->buy_price = 50;
        doubleRate->sell_price = 20;
        doubleRate->affect[0] = 0.5;
        doubleRate->name = "Double rate";
        doubleRate->type = TURRET_MOD_RATE;
    
    game->buyTurret( rocketLauncher , 1 , 2 );
    game->buyTurret( doubleRate , 1 , 3 );
    
    game->go( waveRunLogicManyInRangeCalls , mobs , defaultMobInRangeForManyCalls , defaultMobFellDead , defaultMobReachedTheEnd , NULL );
    
    /**
     * Clear memory
     */
    for( int i = 0 ; i < (int) mobs.size() ; i++ )
        delete mobs[i];
    
    delete spider;
    delete game;
}

void smarterTest(){
    const char INPUT_FILE_NAME[] = "input.in";
    const char OUTPUT_FILE_NAME[] = "output.out";
    
    static int got_helper[MAX_TURRET_COUNT];
    
    ManyRunsLogicInfo *logicInfo = new ManyRunsLogicInfo;
    
    TextGame *game = TextGameReader::read( INPUT_FILE_NAME );
    FILE *out = fopen( OUTPUT_FILE_NAME , "r" );
    
    eprintf("Input %sOK", (game ? "" : "not "));
    
    std::vector< FieldCellCoord > upgrades;
    
    for( int w = 0 ; w < (int) game->all_waves.size() ; w++ ){
        int instructions;
        
        MY_CHECK_FEOF( out );
                            
        if( fscanf( out , "%d" , &instructions ) == 1 ){
            if( instructions < 0 or instructions >= MAX_INSTRUCTIONS_PER_WAVE ){
                MY_ERROR_DIE( 0 , "Invalid number of instructions" );
            }
            
            for( int i = 0 ; i < instructions ; i++ ){
                int type;
                
                MY_CHECK_FEOF( out );
                            
                if( fscanf( out , "%d" , &type ) != 1 ){
                    MY_ERROR_DIE( 0 , "Invalid format" );
                }
                
                if( type < 1 or type == INSTRUCTIONS_REQUIRES_COORDS or type > LAST_VALID_INSTRUCTION_ID ){
                    eprintf("Invalid instruction (#%d:%d) type %d", w + 1, i, type);
                
                    MY_ERROR_DIE( 0 , "Invalid instruction" );
                }
                
                if( type < INSTRUCTIONS_REQUIRES_COORDS ){
                    if( type == INSTRUCTION_ID_SET_ORDER ){
                        int for_type;
                    
                        if( fscanf( out , "%d" , &for_type ) != 1 ){
                            MY_ERROR_DIE( 0 , "Invalid format" );
                        }
                        
                        if( for_type < 0 or for_type >= MAX_MOB_TYPES ){
                            MY_ERROR_DIE( 0 , "Invalid type for order setting" );
                        }
                        
                        std::vector< int > order;
                        
                        int turret_x, turret_y;
                        
                        MY_CHECK_FEOF( out );
                        
                        if( fscanf( out , "%d" , &turret_x ) != 1 ){
                            MY_ERROR_DIE( 0 , "Invalid format" );
                        }
                        
                        if( turret_x >= 0 ){
                            if( game->totalTurretsBuild() == 0 ){
                                /// nonsense
                                MY_ERROR_DIE( 0 , "Trying to set order when no turrets are built" );
                            }
                            
                            if( turret_x >= MAX_TURRET_COUNT ){
                                MY_ERROR_DIE( 0 , "Invalid turret id for order setting" );
                            }
                            
                            if( fscanf( out , "%d" , &turret_y ) != 1 ){
                                MY_ERROR_DIE( 0 , "Invalid format" );
                            }
                            
                            if( !game->field->isInside( turret_x , turret_y ) ){
                                MY_ERROR_DIE( 0 , "Coordinates out of range" );
                            }
                            
                            order.push_back( game->field->getTurretId( turret_x , turret_y ) );
                            
                            for( int j = 0 ; j < game->totalTurretsBuild() - 1 ; j++ ){
                                MY_CHECK_FEOF( out );
                                
                                if( fscanf( out , "%d %d" , &turret_x, &turret_y ) != 2 ){
                                    MY_ERROR_DIE( 0 , "Invalid format" );
                                }
                                
                                if( !game->field->isInside( turret_x , turret_y ) ){
                                    MY_ERROR_DIE( 0 , "Coordinates out of range" );
                                }
                                
                                order.push_back( game->field->getTurretId( turret_x , turret_y ) );
                            }
                            
                            /// Fill the permutation...
                            std::fill( got_helper , got_helper + MAX_TURRET_COUNT , 0 );
                            
                            for( int j = 0 ; j < game->totalTurretsBuild() ; j++ ){
                                if( got_helper[ order[j] ] ){
                                    MY_ERROR_DIE( 0 , "Same turret appears twice in an ordering" );
                                }
                                
                                got_helper[ order[j] ]++;
                            }
                            
                            for( int j = 0 ; j < MAX_TURRET_COUNT ; j++ )
                                if( got_helper[j] == 0 )
                                    order.push_back( j );
                            
                            if( !logicInfo->setManualOrdering( for_type , order ) ){
                                pprintf("Failed to set manual order for %d (not a permutation?)", for_type);
                            
                                MY_ERROR_DIE( 0 , "Setting manual ordering failed" );
                            }else{
                                pprintf("Set manual order for %d", for_type);
                            }
                        }else{
                            pprintf("Returning to default ordering", for_type);
                            
                            logicInfo->isManualOrder[for_type] = false;
                            logicInfo->manualOrder[for_type].clear();
                        }
                    }
                }else{
                    int x, y;
                    
                    MY_CHECK_FEOF( out );
                            
                    if( fscanf( out , "%d %d", &x, &y ) != 2 ){
                        MY_ERROR_DIE( 0 , "Invalid format" );
                    }
                    
                    if( !game->field->isInside( x , y ) ){
                        MY_ERROR_DIE( 0 , "Coordinates out of range" );
                    }
                    
                    if( type == INSTRUCTION_ID_SET_MAX_WASTE ){
                        float waste;
                        
                        MY_CHECK_FEOF( out );
                    
                        if( fscanf( out , "%f" , &waste ) != 1 ){
                            MY_ERROR_DIE( 0 , "Invalid format" );
                        }
                        
                        if( !game->field->isTurret( x , y ) ){
                            MY_ERROR_DIE( 0 , "Setting max waste on path / blocked / free cell" );
                        }
                        
                        logicInfo->maxWasteDamage[x][y] = waste;
                        
                        pprintf("Set max. waste of (%d,%d) to %.3f", x, y, waste);
                    }else if( type == INSTRUCTION_ID_SELL ){
                        if( !game->sellTurret( x ,  y ) ){
                            MY_ERROR_DIE( 0 , "Selling turret failed" );
                        }
                        
                        pprintf("Sold turret at (%d,%d)", x, y);
                    }else if( type == INSTRUCTION_ID_BUY ){
                        int turret_id;
                        
                        MY_CHECK_FEOF( out );
                        
                        if( fscanf( out , "%d", &turret_id ) != 1 ){
                            MY_ERROR_DIE( 0 , "Invalid format" );
                        }
                        
                        if( turret_id < 0 or turret_id >= (int) game->all_turrets.size() ){
                            MY_ERROR_DIE( 0 , "Invalid turret id" );
                        }
                        
                        int needed = game->upgrade_graph[turret_id].size();
                        
                        std::vector< FieldCellCoord > upgrades;
                        
                        for( int j = 0 ; j < needed ; j++ ){
                            int upgrade_x, upgrade_y;
                            
                            MY_CHECK_FEOF( out );
                            
                            if( fscanf( out , "%d %d", &upgrade_x, &upgrade_y ) != 2 ){
                                MY_ERROR_DIE( 0 , "Invalid format" );
                            }
                            
                            if( !game->field->isInside( upgrade_x , upgrade_y )
                               or !game->field->isTurret( upgrade_x , upgrade_y ) ){
                                MY_ERROR_DIE( 0 , "Turret upgrade coords out-of-range / not turret's coords");
                            }
                            
                            upgrades.push_back( FieldCellCoord( upgrade_x , upgrade_y ) );
                        }
                        
                        if( game->buyTurret( turret_id , x , y , upgrades ) < 0 ){
                            MY_ERROR_DIE( 0 , "Failed to buy turret" );
                        }else{
                            /// Set max waste to each new tower to 1.0
                            logicInfo->maxWasteDamage[x][y] = 1.0;
                            
                            pprintf("Bought turret %d at (%d,%d) for %d", (int) turret_id, (int) x, (int) y, (int) game->all_turrets[turret_id]->buy_price);
                        }
                    }
                }
            }
        }else{
            MY_ERROR_DIE( 0 , "Invalid format" );
        }
        
        int len = game->all_waves[w].size();
        
        std::vector< RunningMob* > mobs;
        
        for( int i = 0 ; i < len ; i++ ){
            int type = game->all_waves[w][i].first;
            int count = game->all_waves[w][i].second;
            
            for( int j = 0 ; j < count ; j++ )
                mobs.push_back( new RunningMob( game->all_mobs[type] ) );
        }
        
        game->go( waveRunLogicManyInRangeCalls ,
                  mobs , 
                  defaultMobInRangeForManyCalls ,
                  defaultMobFellDead , 
                  defaultMobReachedTheEnd ,
                  logicInfo
                );
        
        if( game->lives <= 0 ){
            pprintf("Died at wave #%d", w + 1);
            
            MY_OK_DIE( w , "Killed" );
        }else{
            game->money += game->money_on_completion[w];

            pprintf("Survived wave %d, got %" FORMAT_int64 " has %" FORMAT_int64, w + 1, game->money_on_completion[w], game->money);
        }
    }
    
    fclose( out );
    
    pprintf("Survived all waves with %d live(s)", game->lives);
    
    MY_OK_DIE( (int) game->all_waves.size() , "Survived all waves" );
}

void go(){
    //stupidTest();
    
    smarterTest();
}

int main(){
    //freopen( "error.err" , "w" , stderr );
    
    go();
    
    return 0;
}
